% (April 19, 2016)
% 
% Author:
% Jordan Frecon (jordan.frecon@ens-lyon.fr) 
% --> Contact author Nelly Pustelnik (nelly.pustelnik@ens-lyon.fr)
% 
% Contributors:
% Nelly Pustelnik (nelly.pustelnik@ens-lyon.fr)
% Patrice Abry (patrice.abry@ens-lyon.fr)
% Laurent Condat (laurent.condat@gipsa-lab.grenoble-inp.fr)
%
% This software is governed by the CeCILL license under French law and
% abiding by the rules of distribution of free software.  You can  use,
% modify and/ or redistribute the software under the terms of the CeCILL
% license as circulated by CEA, CNRS and INRIA at the following URL
% "http://www.cecill.info".
% 
% As a counterpart to the access to the source code and  rights to copy,
% modify and redistribute granted by the license, users are provided only
% with a limited warranty  and the software's author,  the holder of the
% economic rights,  and the successive licensors  have only  limited
% liability.
% 
% In this respect, the user's attention is drawn to the risks associated
% with loading,  using,  modifying and/or developing or reproducing the
% software by the user in light of its specific status of free software,
% that may mean  that it is complicated to manipulate,  and  that  also
% therefore means  that it is reserved for developers  and  experienced
% professionals having in-depth computer knowledge. Users are therefore
% encouraged to load and test the software's suitability as regards their
% requirements in conditions enabling the security of their systems and/or
% data to be ensured and,  more generally, to use and operate it in the
% same conditions as regards security.
% 
% The fact that you are presently reading this means that you have had
% knowledge of the CeCILL license and that you accept its terms.
%
%--------------------------------------------------------------------------
% TV On-The-Fly: On-the-fly approximation of the multivariate TV problem  
%                                                                         
% For theoretical aspects please refer to :                               
% J. Frecon, N. Pustelnik, P. Abry, L. Condat                             
% On-The-Fly Approximation of Multivariate Total Variation Minimization   
% IEEE Transactions on Signal Processing,                                 
% Vol. 64, Issue 9, pp. 2355-2364, May. 2016                              
%--------------------------------------------------------------------------
%
% [ X_otf ] = TV_OnTheFly_stream( Y, Lambda, Q, X )
%
% Computes an on-the-fly approximate solution 'X_otf' to the multivariate
% total variation (TV) problem applied to the multivariate observation 'Y'.
% Shows the on-the-fly solution as the data streams incomes.
% Note: displaying the solution at each step may be very costly.
%
% Input:
%   - 'Y' Multivariate observation
%   - 'Lambda' TV regularization parameter
%   - 'Q' Quality parameter for the approximated algorithm
%   - 'param' [optional]
%       - 'param.X' true signal
%       - 'param.fps' number of frames per second
%
% Output:
%   - 'X_otf' Solution of the proposed algorithm
%
% Dependency:
%   - TV_OnTheFly_aux
%       - changePointDetection
%       - changePointDetection_bound
%       - zetaEstimation            
%
function [ X_otf ] = TV_OnTheFly_stream( Y, Lambda, Q, param )





if nargin <= 3
    param = struct;
end

if ~isfield(param,'X');
    param.X = NaN*ones(size(Y));
end

if ~isfield(param,'fps');
    param.fps = 2^8;
end

X   = param.X;
fps = param.fps;


% Initialize algorithm
N     = length(Y);
M     = size(Y,1);
Init  = 0;
X_otf = [];
k0    = 1;

if size(Y,1) == 1
    Init = 1;
end

[temp, aux, k0,kOut] = TV_OnTheFly_aux(Y(:,k0), Lambda, Q, k0, k0, Init);
aux.sigma2           = ones(Q,1)*var(Y,[],2)';




% Plots options
grey    = .6.*[1 1 1];
red     = [1 0 0];
black   = [0 0 0];
orange  = [0.929 0.694 0.125];
blue    = [0 0.447 0.741];
markers = {grey,black,red,orange,blue};
options = {{},{'LineWidth', 3},{'LineWidth', 3,'LineStyle','-.'},{},{}};


% Initialize figure
Fig=figure;
clf
for isp = 1 : M
    ax(isp) = subplot (M, 1, isp);
    hold on;
    for ipl = 1 : length(markers)
        hp(isp, ipl) = plot (NaN, NaN, 'color',markers{ipl}, options{ipl}{:});
    end
    axis(ax(isp),[0 N min(Y(isp,:)) max(Y(isp,:))]);
    hold off;
    grid on;
    
    if isp==1
        title('Online result');
    end
end
refreshdata(Fig,'caller');
drawnow;
xx = 1:N;





timer = tic;
for(k=2:N)
   
    
    
    kp = kOut; t_temp=0;
    while(kp <= k);
        

        % - Temporary solution (up to kp) with boundary conditions
        k_bnd       = kp;
        k0_bnd      = k0;
        aux_bnd     = aux;
        %X_otf_bnd   = X_otf;
        X_otf_bnd = NaN*ones(size(Y));
        if ~isempty(X_otf)
            X_otf_bnd(:,1:size(X_otf,2))=X_otf;
        end
        
        while(k_bnd <= kp)
            
            [temp_bnd, aux_bnd, k0_bnd, k_bnd] = TV_OnTheFly_aux(Y(:,k_bnd), Lambda, Q, k_bnd, k0_bnd, Init, aux_bnd, kp);
            %X_otf_bnd = horzcat(X_otf_bnd, temp_bnd);
            if ~isempty(temp_bnd)
            X_otf_bnd(:,1+size(X_otf,2):size(X_otf,2)+size(temp_bnd,2))=temp_bnd;
            end
        end
        

        % - Approximated solution (up to kp)
        tstart                  = tic;
        [temp, aux, k0,kOut]    = TV_OnTheFly_aux(Y(:,kp), Lambda, Q, kp, k0, Init, aux, N);
        X_otf                   = horzcat(X_otf, temp);
        t_temp                  = t_temp+toc(tstart);
        
        
        
        % - Display
        for isp = 1 : M
            
            hp(isp, 1).XDataSource = 'xx(1:k)';
            hp(isp, 1).YDataSource = 'Y(isp,1:k)';
            
            hp(isp, 2).XDataSource = 'xx(1:k)';
            hp(isp, 2).YDataSource = 'X(isp,1:k)';
            
            hp(isp, 3).XDataSource = 'xx(1:length(X_otf_bnd))';
            hp(isp, 3).YDataSource = 'X_otf_bnd(isp, :)';
           
            hp(isp, 4).XDataSource = '[kp kp]';
            hp(isp, 4).YDataSource = 'ylim(ax(isp))';
            
            hp(isp, 5).XDataSource = '[k0 k0]';
            hp(isp, 5).YDataSource = 'ylim(ax(isp))';
             
        end


        elap = toc (timer);
        if elap >= 1 / fps
            for isp=1:M
                for ilm=1:5
                refreshdata(hp(isp,ilm),'caller');
                end
            end
            drawnow limitrate
            timer = tic;
        end
        

        
        
        kp = kOut;
        
    end
    
    
end





end

